package dk.vajhoej.record.test;

import static org.junit.Assert.*;

import java.math.BigDecimal;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Map;
import java.util.TimeZone;

import org.junit.Test;

import dk.vajhoej.record.CodeGen;
import dk.vajhoej.record.CompositFastStructReader;
import dk.vajhoej.record.RecordException;
import dk.vajhoej.record.FastStructReader;
import dk.vajhoej.record.StructInfoCache;
import dk.vajhoej.record.StructReader;

public class TestFastStructReader {
	private final static int N = 100000;
	@Test
	public void testReadSimple() {
		byte[] b = { 0x02, 0x01, 0x00, 0x00,
			         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
				     0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
		try {
			FastStructReader<Data> fsr = CodeGen.genReader(Data.class);
			fsr.init(b);
			Data o = fsr.read(Data.class);
			assertEquals("iv", 258, o.getIv());
			assertEquals("xv", 123.456, o.getXv(), 0.0005);
			assertEquals("sv", "ABC     ", o.getSv());
            assertEquals("more", false, fsr.more());
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
	}
	@Test
	public void testReadMulti() {
		byte[] b = { 0x01, 0x00, 0x00, 0x00,
				     0x01, 0x00, 0x00, 0x00,
			         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
			         0x03, 0x00, 0x00, 0x00,
				     0x02, 0x00, 0x00, 0x00,
				     0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20 };
		try {
			Map<Class<? extends SuperData>, FastStructReader<? extends SuperData>> map = new HashMap<Class<? extends SuperData>, FastStructReader<? extends SuperData>>();
			map.put(SubDataOne.class, CodeGen.genReader(SubDataOne.class));
			map.put(SubDataTwo.class, CodeGen.genReader(SubDataTwo.class));
			FastStructReader<SuperData> fsr = new CompositFastStructReader<SuperData>(SuperData.class, map);
			fsr.init(b);
			SuperData o1 = fsr.read(SuperData.class);
			assertEquals("id", 1, o1.getId());
			assertEquals("typ", 1, o1.getTyp());
			SubDataOne o1m = (SubDataOne)o1;
			assertEquals("x", 123.456, o1m.getX(), 0.0005);
			SuperData o2 = fsr.read(SuperData.class);
			assertEquals("id", 3, o2.getId());
			assertEquals("typ", 2, o2.getTyp());
			SubDataTwo o2m = (SubDataTwo)o2;
			assertEquals("s", "ABC         ", o2m.getS());
            assertEquals("more", false, fsr.more());
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
	}
    @Test
    public void testReadMultiAlt() {
        byte[] b = { 0x01, 0x00, 0x00, 0x00,
                     0x01, 0x00, 0x00, 0x00,
                     0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                     0x03, 0x00, 0x00, 0x00,
                     0x02, 0x00, 0x00, 0x00,
                     0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20 };
        try {
			FastStructReader<SubDataOne> fsr1 = CodeGen.genReader(SubDataOne.class);
			fsr1.init(b);
            SubDataOne o1m = fsr1.read(SubDataOne.class);
            assertEquals("id", 1, o1m.getId());
            assertEquals("typ", 1, o1m.getTyp());
            assertEquals("x", 123.456, o1m.getX(), 0.0005);
            assertEquals("more", true, fsr1.more());
			FastStructReader<SubDataTwo> fsr2 = CodeGen.genReader(SubDataTwo.class);
			fsr2.init(fsr1.getByteBuffer());
            SubDataTwo o2m = fsr2.read(SubDataTwo.class);
            assertEquals("id", 3, o2m.getId());
            assertEquals("typ", 2, o2m.getTyp());
            assertEquals("s", "ABC         ", o2m.getS());
            assertEquals("more", false, fsr2.more());
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testReadMultiPad() {
        byte[] b = { 0x01, 0x00, 0x00, 0x00,
                     0x01, 0x00, 0x00, 0x00,
                     0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40, 0x00, 0x00, 0x00, 0x00,
                     0x03, 0x00, 0x00, 0x00,
                     0x02, 0x00, 0x00, 0x00,
                     0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20 };
        try {
			Map<Class<? extends SuperDataPad>, FastStructReader<? extends SuperDataPad>> map = new HashMap<Class<? extends SuperDataPad>, FastStructReader<? extends SuperDataPad>>();
			map.put(SubDataOnePad.class, CodeGen.genReader(SubDataOnePad.class));
			map.put(SubDataTwoPad.class, CodeGen.genReader(SubDataTwoPad.class));
			FastStructReader<SuperDataPad> fsr = new CompositFastStructReader<SuperDataPad>(SuperDataPad.class, map);
			fsr.init(b);
            SuperDataPad o1 = fsr.read(SuperDataPad.class);
            assertEquals("id", 1, o1.getId());
            assertEquals("typ", 1, o1.getTyp());
            SubDataOnePad o1m = (SubDataOnePad)o1;
            assertEquals("x", 123.456, o1m.getX(), 0.0005);
            SuperDataPad o2 = fsr.read(SuperDataPad.class);
            assertEquals("id", 3, o2.getId());
            assertEquals("typ", 2, o2.getTyp());
            SubDataTwoPad o2m = (SubDataTwoPad)o2;
            assertEquals("s", "ABC         ", o2m.getS());
            assertEquals("more", false, fsr.more());
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
	@Test
	public void testReadPerf() {
		byte[] b = { 0x02, 0x01, 0x00, 0x00,
			         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
				     0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
		byte[] bm = new byte[N*b.length];
		for(int i = 0; i < N; i++) {
			System.arraycopy(b, 0, bm, i*b.length, b.length);
		}
		try {
			long t1 = System.currentTimeMillis();
			FastStructReader<Data> fsr = CodeGen.genReader(Data.class);
			fsr.init(bm);
			for(int i = 0; i < N; i++) {
				Data o = fsr.read(Data.class);
				assertEquals("iv #" + i, 258, o.getIv());
			}
			long t2 = System.currentTimeMillis();
			assertTrue("time", t2-t1 < 0.001*N);
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
	}
	@Test
	public void testReadRelPerf() {
		byte[] b = { 0x02, 0x01, 0x00, 0x00,
			         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
				     0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
		byte[] bm = new byte[N*b.length];
		for(int i = 0; i < N; i++) {
			System.arraycopy(b, 0, bm, i*b.length, b.length);
		}
		try {
			long t1 = System.currentTimeMillis();
			StructReader sr = new StructReader(bm);
			for(int i = 0; i < N; i++) {
				Data o = sr.read(Data.class);
			}
			long t2 = System.currentTimeMillis();
			assertEquals("more", false, sr.more());
			long t1f = System.currentTimeMillis();
			FastStructReader<Data> fsr = CodeGen.genReader(Data.class);
			fsr.init(bm);
			for(int i = 0; i < N; i++) {
				Data o = fsr.read(Data.class);
			}
			long t2f = System.currentTimeMillis();
			assertEquals("more", false, fsr.more());
			assertTrue("relative time", (t2f - t1f) * 1.0 / (t2 - t1) < 0.2);
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
	}
	@Test
	public void testReadCache() {
		byte[] b = { 0x02, 0x01, 0x00, 0x00,
			         0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
				     0x41, 0x42, 0x43, 0x20, 0x20, 0x20, 0x20, 0x20 };
		try {
			StructInfoCache.getInstance().reset();
			FastStructReader<Data> fsr = CodeGen.genReader(Data.class);
			fsr.init(b);
			Data o = fsr.read(Data.class);
			assertEquals("iv", 258, o.getIv());
			assertTrue("hit rate", StructInfoCache.getInstance().getHitRate() < 0.001);
			for(int i = 0; i < 9; i++) {
				fsr.init(b);
				Data o2 = fsr.read(Data.class);
				assertEquals("iv", 258, o2.getIv());
			}
			assertTrue("hit rate", StructInfoCache.getInstance().getHitRate() < 0.001);
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
	}
    @Test
    public void testReadBoolean() {
        byte[] b = { 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01 };
        try {
			FastStructReader<BooleanData> fsr = CodeGen.genReader(BooleanData.class);
			fsr.init(b);
            BooleanData o = fsr.read(BooleanData.class);
            assertEquals("b1", true, o.isB1());
            assertEquals("b2", false, o.isB2());
            assertEquals("b3", true, o.isB3());
            assertEquals("more", false, fsr.more());
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testReadBits() {
        byte[] b = { 0x01, 0x23, 0x04, (byte)0x80, 0x02 };
        try {
			FastStructReader<BitData> fsr = CodeGen.genReader(BitData.class);
			fsr.init(b);
            BitData o = fsr.read(BitData.class);
            assertEquals("i1", 1, o.getI1());
            assertEquals("n1", 2, o.getN1());
            assertEquals("n2", 3, o.getN2());
            assertEquals("i3", 4, o.getI2());
            assertEquals("z1", 1, o.getZ1());
            assertEquals("z2", 2, o.getZ2());
            assertEquals("more", false, fsr.more());
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
	@Test
	public void testReadTime() {
		byte[] b = { 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
				     0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
				     0x01, 0x00, 0x00, 0x00,
				     0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
		try {
			FastStructReader<TimeData> fsr = CodeGen.genReader(TimeData.class);
			fsr.init(b);
			TimeData o = fsr.read(TimeData.class);
			Calendar cal = new GregorianCalendar(TimeZone.getTimeZone("GMT"));
			cal.setTime(o.getT1());
			assertEquals("t1 year", 1970, cal.get(Calendar.YEAR));
			assertEquals("t1 month", Calendar.JANUARY, cal.get(Calendar.MONTH));
			assertEquals("t1 day", 1, cal.get(Calendar.DAY_OF_MONTH));
			assertEquals("t1 hour", 0, cal.get(Calendar.HOUR_OF_DAY));
			assertEquals("t1 minute", 0, cal.get(Calendar.MINUTE));
			assertEquals("t1 second", 0, cal.get(Calendar.SECOND));
			assertEquals("t1 millisecond", 1, cal.get(Calendar.MILLISECOND));
			cal.setTime(o.getT2());
			assertEquals("t2 year", 1858, cal.get(Calendar.YEAR));
			assertEquals("t2 month", Calendar.NOVEMBER, cal.get(Calendar.MONTH));
			assertEquals("t2 day", 17, cal.get(Calendar.DAY_OF_MONTH));
			assertEquals("t2 hour", 4, cal.get(Calendar.HOUR_OF_DAY));
			assertEquals("t2 minute", 0, cal.get(Calendar.MINUTE));
			assertEquals("t2 second", 0, cal.get(Calendar.SECOND));
			assertEquals("t2 millisecond", 0, cal.get(Calendar.MILLISECOND));
			cal.setTime(o.getT3());
			assertEquals("t3 year", 1970, cal.get(Calendar.YEAR));
			assertEquals("t3 month", Calendar.JANUARY, cal.get(Calendar.MONTH));
			assertEquals("t3 day", 1, cal.get(Calendar.DAY_OF_MONTH));
			assertEquals("t3 hour", 0, cal.get(Calendar.HOUR_OF_DAY));
			assertEquals("t3 minute", 0, cal.get(Calendar.MINUTE));
			assertEquals("t3 second", 1, cal.get(Calendar.SECOND));
			assertEquals("t3 millisecond", 0, cal.get(Calendar.MILLISECOND));
			cal.setTime(o.getT4());
			assertEquals("t4 year", 1970, cal.get(Calendar.YEAR));
			assertEquals("t4 month", Calendar.JANUARY, cal.get(Calendar.MONTH));
			assertEquals("t4 day", 1, cal.get(Calendar.DAY_OF_MONTH));
			assertEquals("t4 hour", 0, cal.get(Calendar.HOUR_OF_DAY));
			assertEquals("t4 minute", 0, cal.get(Calendar.MINUTE));
			assertEquals("t4 second", 1, cal.get(Calendar.SECOND));
			assertEquals("t4 millisecond", 0, cal.get(Calendar.MILLISECOND));
            assertEquals("more", false, fsr.more());
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		}
	}
    @Test
    public void testReadBCD() {
        byte[] b = { 0x12, 0x34, 0x56, 0x7D,
                     (byte)0xF1, (byte)0xF2, (byte)0xF7, (byte)0xF9, (byte)0xF5, (byte)0xC0  };
        try {
			FastStructReader<BCDData> fsr = CodeGen.genReader(BCDData.class);
			fsr.init(b);
            BCDData o = fsr.read(BCDData.class);
            assertEquals("v1", new BigDecimal("-12345.67"), o.getV1());
            assertEquals("v2", new BigDecimal("1279.50"), o.getV2());
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testReadVaxFloat() {
        byte[] b = { 0x45, 0x42, (byte)0xA4, 0x70,
                     0x7E, 0x40, 0x2F, (byte)0xDD, (byte)0x9F, 0x1A, 0x77, (byte)0xBE };
        try {
			FastStructReader<VaxFloatData> fsr = CodeGen.genReader(VaxFloatData.class);
			fsr.init(b);
            VaxFloatData o = fsr.read(VaxFloatData.class);
            assertEquals("v1", 12.34f, o.getV1(), 0.0001f);
            assertEquals("v2", 123.456, o.getV2(), 0.00001);
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testReadVaxFloatArray() {
        byte[] b = { (byte)0x99, 0x40, (byte)0x9A, (byte)0x99,
        		     0x45, 0x42, (byte)0xA4, 0x70,
        		     (byte)0xF6, 0x43, 0x79, (byte)0xE9,
        		     0x00, 0x00, 0x00, 0x00,
        		     0x00, (byte)0xC0, 0x00, 0x00,
        		     0x20, (byte)0xC1, 0x00, 0x00,
        		     (byte)0xDB, 0x2F, (byte)0xFF, (byte)0xE6,
        		     0x15, 0x51, (byte)0xF9, 0x02,
        		     0x13, 0x40, 0x33, 0x33, 0x33, 0x33, 0x33, 0x33,
        		     0x48, 0x40, 0x14, (byte)0xAE, (byte)0xE1, 0x7A, (byte)0xAE, 0x47,
        		     0x7E, 0x40, 0x2F, (byte)0xDD, (byte)0x9F, 0x1A, 0x77, (byte)0xBE,
        		     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        		     0x00, (byte)0xC0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        		     0x24, (byte)0xC0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        		     (byte)0xFB, 0x3D, (byte)0xDF, 0x7C, (byte)0xD7, (byte)0xD9, (byte)0xBB, (byte)0xBD,
        		     0x22, 0x42, 0x5F, (byte)0xA0, 0x00, 0x20, 0x00, 0x00,
        		     (byte)0x99, 0x40, (byte)0x99, (byte)0x99, (byte)0x99, (byte)0x99, (byte)0x98, (byte)0x99,
        		     0x45, 0x42, (byte)0xA3, 0x70, 0x0A, (byte)0xD7, 0x70, 0x3D,
        		     (byte)0xF6, 0x43, 0x78, (byte)0xE9, (byte)0xFD, (byte)0xD4, (byte)0xB8, (byte)0xF3,
        		     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        		     0x00, (byte)0xC0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        		     0x20, (byte)0xC1, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
        		     (byte)0xDB, 0x2F, (byte)0xFE, (byte)0xE6, (byte)0xBD, (byte)0xCE, (byte)0xD8, (byte)0xED,
        		     0x15, 0x51, (byte)0xF9, 0x02, 0x00, 0x00, 0x00, 0x00 };
        float[] x1 = { 1.2f, 12.34f, 123.456f, 0.0f, -0.5f, -2.5f, 0.0000000001f, 10000000000.0f };
        double[] x2 = { 1.2, 12.34, 123.456, 0.0, -0.5, -2.5, 0.0000000001, 10000000000.0 };
        double[] x3 = { 1.2, 12.34, 123.456, 0.0, -0.5, -2.5, 0.0000000001, 10000000000.0 };
        try {
			FastStructReader<VaxFloatArrayData> fsr = CodeGen.genReader(VaxFloatArrayData.class);
			fsr.init(b);
            VaxFloatArrayData o = fsr.read(VaxFloatArrayData.class);
            float[] v1 = o.getV1();
            double[] v2 = o.getV2();
            double[] v3 = o.getV3();
            for(int i = 0; i < 8; i++) {
                assertEquals("v1[" + i + "]", x1[i], v1[i], Math.max(x1[i] * 0.0000001f, 0.1e-30f));
                assertEquals("v2[" + i + "]", x2[i], v2[i], Math.max(x2[i] * 0.00000001, 0.1e-30));
                assertEquals("v3[" + i + "]", x2[i], v3[i], Math.max(x3[i] * 0.00000001, 0.1e-30));
            }
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testReadUnsigned() {
        byte[] b = { (byte)0xFF,
                     (byte)0xFF, (byte)0xFF,
                     (byte)0xFF, (byte)0xFF,(byte)0xFF, (byte)0xFF };
        try {
			FastStructReader<UnsignedData> fsr = CodeGen.genReader(UnsignedData.class);
			fsr.init(b);
            UnsignedData o = fsr.read(UnsignedData.class);
            assertEquals("ui1v", 255, o.getUi1v());
            assertEquals("ui2v", 65535, o.getUi2v());
            assertEquals("ui4v", 4294967295L, o.getUi4v());
            assertEquals("more", false, fsr.more());
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testReadArray() {
        byte[] b = { 0x02, 0x01, 0x00, 0x00,
                     0x03, 0x01, 0x00, 0x00,
                     0x04, 0x01, 0x00, 0x00,
                     0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                     0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                     0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40 };
        try {
			FastStructReader<ArrayData> fsr = CodeGen.genReader(ArrayData.class);
			fsr.init(b);
            ArrayData o = fsr.read(ArrayData.class);
            for(int ix = 0; ix < 3; ix++) {
                assertEquals("iv", 258 + ix, o.getIv(ix));
                assertEquals("xv", 123.456, o.getXv(ix), 0.0005);
            }
            assertEquals("more", false, fsr.more());
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testReadCountArray() {
        byte[] b = { 0x04,
        		     0x02, 0x01, 0x00, 0x00,
                     0x03, 0x01, 0x00, 0x00,
                     0x04, 0x01, 0x00, 0x00,
                     0x05, 0x01, 0x00, 0x00,
                     0x04, 0x00,
                     0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                     0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                     0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40,
                     0x77, (byte)0xBE, (byte)0x9F, 0x1A, 0x2F, (byte)0xDD, 0x5E, 0x40 };
        try {
			FastStructReader<CountArrayData> fsr = CodeGen.genReader(CountArrayData.class);
			fsr.init(b);
            CountArrayData o = fsr.read(CountArrayData.class);
            for(int ix = 0; ix < 4; ix++) {
                assertEquals("iv", 258 + ix, o.getIv(ix));
                assertEquals("xv", 123.456, o.getXv(ix), 0.0005);
            }
            assertEquals("more", false, fsr.more());
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
	@Test
	public void testReadStructField() {
		byte[] b = { 0x01, 0x00, 0x00, 0x00,
     			     0x02, 0x00, 0x00, 0x00,
	    		     0x03, 0x00, 0x00, 0x00,
	    		     0x04, 0x00, 0x00, 0x00,
	    		     0x05, 0x00, 0x00, 0x00 };
		try {
			FastStructReader<MainData> fsr = CodeGen.genReader(MainData.class);
			fsr.init(b);
			@SuppressWarnings("unused")
			MainData o = fsr.read(MainData.class);
			fail("Struct fields not supported");
		} catch (RecordException e) {
		}
	}
    @Test
    public void testReadStructAlignment0() {
        byte[] b = { 0x01, 0x02, 0x00, 0x03,
                     0x00, 0x00, 0x00, 0x04,
                     0x00, 0x00, 0x00, 0x00,
                     0x00, 0x00, 0x00 };
        try {
			FastStructReader<AlignData0> fsr = CodeGen.genReader(AlignData0.class);
			fsr.init(b);
            AlignData0 o = fsr.read(AlignData0.class);
            assertEquals("i1", 1, o.getI1());
            assertEquals("i2", 2, o.getI2());
            assertEquals("i4", 3, o.getI4());
            assertEquals("i8", 4, o.getI8());
            assertEquals("more", false, fsr.more());
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testReadStructAlignment1() {
        byte[] b = { 0x01, 0x00, 0x02, 0x00,
                     0x03, 0x00, 0x00, 0x00,
                     0x04, 0x00, 0x00, 0x00,
                     0x00, 0x00, 0x00, 0x00 };
        try {
			FastStructReader<AlignData1> fsr = CodeGen.genReader(AlignData1.class);
			fsr.init(b);
            AlignData1 o = fsr.read(AlignData1.class);
            assertEquals("i1", 1, o.getI1());
            assertEquals("i2", 2, o.getI2());
            assertEquals("i4", 3, o.getI4());
            assertEquals("i8", 4, o.getI8());
            assertEquals("more", false, fsr.more());
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testReadStructAlignment2() {
        byte[] b = { 0x01, 0x00, 0x00, 0x00,
                     0x00, 0x00, 0x00, 0x00,
                     0x02, 0x00, 0x00, 0x00,
                     0x00, 0x00, 0x00, 0x00,
                     0x03, 0x00, 0x00, 0x00,
                     0x00, 0x00, 0x00, 0x00,
                     0x04, 0x00, 0x00, 0x00,
                     0x00, 0x00, 0x00, 0x00 };
        try {
			FastStructReader<AlignData2> fsr = CodeGen.genReader(AlignData2.class);
			fsr.init(b);
            AlignData2 o = fsr.read(AlignData2.class);
            assertEquals("i1", 1, o.getI1());
            assertEquals("i2", 2, o.getI2());
            assertEquals("i4", 3, o.getI4());
            assertEquals("i8", 4, o.getI8());
            assertEquals("more", false, fsr.more());
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
 	@Test
	public void testReadError1() {
		byte[] b = { 0x00, 0x00, 0x00, 0x00 };
		try {
			FastStructReader<ErrorData1> fsr = CodeGen.genReader(ErrorData1.class);
			fsr.init(b);
			ErrorData1 o = fsr.read(ErrorData1.class);
			fail("Missing exception");
			o.toString();
		} catch (RecordException e) {
		}
	}
	@Test
	public void testReadError2() {
		byte[] b = { };
		try {
			FastStructReader<ErrorData2> fsr = CodeGen.genReader(ErrorData2.class);
			fsr.init(b);
			ErrorData2 o = fsr.read(ErrorData2.class);
			fail("Missing exception");
			o.toString();
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		} catch (IllegalArgumentException e) {
		}
	}
	@Test
	public void testReadError3() {
		byte[] b = { };
		try {
			FastStructReader<ErrorData3> fsr = CodeGen.genReader(ErrorData3.class);
			fsr.init(b);
			ErrorData3 o = fsr.read(ErrorData3.class);
			fail("Missing exception");
			o.toString();
		} catch (RecordException e) {
			fail("Unexpected exception: " + e);
		} catch (IllegalArgumentException e) {
		}
	}
    @Test
    public void testReadErrorMissing() {
        byte[] b = { };
        try {
			FastStructReader<Data> fsr = CodeGen.genReader(Data.class);
			fsr.init(b);
            Data o = fsr.read(Data.class);
            fail("Missing exception");
            o.toString();
        } catch (RecordException e) {
        }
    }
    @Test
    public void testReadStrings() {
        byte[] b = { 0x41, 0x42, 0x43, 0x44,
                     0x02, 0x00, 0x41, 0x42,
                     0x02, 0x00, 0x41, 0x42, 0x00, 0x00,
                     0x41, 0x42, 0x00, 0x00,
                     0x02, 0x41, 0x42,
                     0x41, 0x42, 0x20, 0x20 };
        try {
			FastStructReader<StringData> fsr = CodeGen.genReader(StringData.class);
			fsr.init(b);
            StringData o = fsr.read(StringData.class);
            assertEquals("s1", "ABCD", o.getS1());
            assertEquals("s2", "AB", o.getS2());
            assertEquals("s3", "AB", o.getS3());
            assertEquals("s4", "AB", o.getS4());
            assertEquals("s5", "AB", o.getS5());
            assertEquals("s6", "AB", o.getS6());
            assertEquals("more", false, fsr.more());
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testReadBigEndian() {
        byte[] b = { 0x00, 0x00, 0x01, 0x02 };
        try {
			FastStructReader<BigEndianData> fsr = CodeGen.genReader(BigEndianData.class);
			fsr.init(b);
            BigEndianData o = fsr.read(BigEndianData.class);
            assertEquals("iv", 258, o.getIv());
            assertEquals("more", false, fsr.more());
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testReadMixedEndian() {
        byte[] b = { 0x00, 0x01, 0x02, 0x00, 0x00, 0x03, 0x00, 0x04 };
        try {
            FastStructReader<MixedEndianData> fsr = CodeGen.genReader(MixedEndianData.class);
            fsr.init(b);
            MixedEndianData o = fsr.read(MixedEndianData.class);
            assertEquals("iv1", 1, o.getIv1());
            assertEquals("iv2", 2, o.getIv2());
            assertEquals("iv3", 3, o.getIv3());
            assertEquals("iv4", 4, o.getIv4());
            assertEquals("more", false, fsr.more());
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testReadEndPad1() {
        byte[] b = { 0x02, 0x01, 0x00, 0x00,
                     0x01 };
        try {
			FastStructReader<EndPadData1> fsr = CodeGen.genReader(EndPadData1.class);
			fsr.init(b);
            EndPadData1 o = fsr.read(EndPadData1.class);
            assertEquals("iv", 258, o.getIv());
            assertEquals("bv", 1, o.getBv());
            assertEquals("more", false, fsr.more());
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testReadEndPad2() {
        byte[] b = { 0x02, 0x01, 0x00, 0x00,
                     0x01, 0x00, 0x00, 0x00 };
        try {
			FastStructReader<EndPadData2> fsr = CodeGen.genReader(EndPadData2.class);
			fsr.init(b);
            EndPadData2 o = fsr.read(EndPadData2.class);
            assertEquals("iv", 258, o.getIv());
            assertEquals("bv", 1, o.getBv());
            assertEquals("more", false, fsr.more());
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void testReadEndPad3() {
        byte[] b = { 0x02, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                     0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
        try {
			FastStructReader<EndPadData3> fsr = CodeGen.genReader(EndPadData3.class);
			fsr.init(b);
            EndPadData3 o = fsr.read(EndPadData3.class);
            assertEquals("iv", 258, o.getIv());
            assertEquals("xv", 0.0, o.getXv(), 0.0005);
            assertEquals("bv", 1, o.getBv());
            assertEquals("more", false, fsr.more());
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void TestReadGeneralInt() {
        byte[] b = { 0x03, 0x02, 0x01, 0x05, 0x04, 0x03, 0x02, 0x01 };
        try {
			FastStructReader<GeneralIntData> fsr = CodeGen.genReader(GeneralIntData.class);
			fsr.init(b);
            GeneralIntData o = fsr.read(GeneralIntData.class);
            assertEquals("v1", 0x0000000000010203L, o.getV1());
            assertEquals("v2", 0x0000000102030405L, o.getV2());
            assertEquals("more", false, fsr.more());
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
    @Test
    public void TestReadRemainingString() {
        byte[] b = { 0x02, 0x01, 0x00, 0x00, 0x41, 0x42, 0x43 };
        try
        {
			FastStructReader<RemainingStringData> fsr = CodeGen.genReader(RemainingStringData.class);
			fsr.init(b);
            RemainingStringData o = fsr.read(RemainingStringData.class);
            assertEquals("iv", 258, o.getIv());
            assertEquals("sv", "ABC", o.getSv());
            assertEquals("more", false, fsr.more());
        } catch (RecordException e) {
            fail("Unexpected exception: " + e);
        }
    }
}
